from collections import OrderedDict

from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response


class RestPageNumberPagination(PageNumberPagination):
    page_size_query_param = "page_size"

    def get_paginated_response(self, data):
        page = self.page

        next_page = page.next_page_number() if page.has_next() else None
        previous_page = page.previous_page_number() if page.has_previous() else None

        return Response(
            OrderedDict(
                [
                    ("count", page.paginator.count),
                    ("num_pages", page.paginator.num_pages),
                    ("next", next_page),
                    ("previous", previous_page),
                    ("page_size", self.get_page_size(self.request)),
                    ("results", data),
                ]
            )
        )



from terra_opp.helpers import Choices

TROPP_URL_FETCHER = "terra_opp.renderers.django_url_fetcher"

TROPP_STATES = Choices(
    ("DRAFT", 100, "Draft"),
    ("SUBMITTED", 200, "Submitted"),
    ("ACCEPTED", 300, "Accepted"),
    ("REFUSED", -1, "Refused"),
    ("CANCELLED", -100, "Cancelled"),
    ("MISSING", 0, "Missing"),
)

TROPP_PICTURES_STATES_WORKFLOW = False

TROPP_VIEWPOINT_PROPERTIES_SET = {
    "pdf": {
        ("camera", "Appareil photo"),
    },
    "form": {},
    "filter": {},
}

TROPP_FEATURES_PROPERTIES_FROM_VIEWPOINT = ()

TROPP_SEARCHABLE_PROPERTIES = {
    "road": {
        "json_key": "voie",
        "type": "text",
    },
    "site": {
        "json_key": "site",
        "type": "text",
    },
}

# Create a geostore layer and provide its Primary Key here
TROPP_OBSERVATORY_LAYER_PK = None



from django.template import Context, Template
from django.test import TestCase

from terra_opp.tests.factories import ViewpointFactory


class AsVersatileTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Create viewpoint with draft picture attached to it
        cls.viewpoint = ViewpointFactory(label="Basic viewpoint")

    def test_image_loader_object(self):
        self.maxDiff = None
        context = Context({"picture": self.viewpoint.pictures.all()[0]})
        template_to_render = Template(
            "{% load opp_tags %}{{ picture.file|as_versatile:'thumbnail__750x1500' }}"
        )
        rendered_template = template_to_render.render(context)
        self.assertIn("750x1500", rendered_template)



from django.apps import AppConfig
from django.conf import settings
from terra_accounts.permissions_mixins import PermissionRegistrationMixin
from rest_framework.reverse import reverse
from django.utils.translation import gettext_lazy as _
from django.db.utils import ProgrammingError


class TerraOppConfig(PermissionRegistrationMixin, AppConfig):
    name = "terra_opp"

    permissions = (
        ("OPP", "can_manage_viewpoints", _("Can manage viewpoints")),
        ("OPP", "can_manage_pictures", _("Can manage pictures")),
    )

    def ready(self):
        from . import checks  # NOQA
        from geostore import models

        super().ready()

        # Set default settings from this app to django.settings if not present
        from . import settings as defaults

        dj_settings = settings._wrapped.__dict__
        for name in dir(defaults):
            dj_settings.setdefault(name, getattr(defaults, name))

        try:
            opp_layer = models.Layer.objects.get(id=settings.TROPP_OBSERVATORY_LAYER_PK)
        except models.Layer.DoesNotExist:
            pass
        except ProgrammingError:
            # Db should not be initialized
            pass
        else:
            terra_settings = getattr(settings, "TERRA_APPLIANCE_SETTINGS", {})
            modules = terra_settings.get("modules", {})

            # Update terra appliance settings with default OPP settings
            modules["OPP"] = {
                "viewpoints": reverse("terra_opp:viewpoint-list"),
                "layer_tilejson": reverse("layer-tilejson", args=(opp_layer.pk,)),
                "searchable_properties": settings.TROPP_SEARCHABLE_PROPERTIES,
                "layerId": opp_layer.pk,
                "layerName": opp_layer.name,
            }
            terra_settings.update({"modules": modules})
            setattr(settings, "TERRA_APPLIANCE_SETTINGS", terra_settings)



from django.core.management import call_command
from django.test import TestCase, override_settings
from io import StringIO

from geostore import GeometryTypes
from geostore.models import Layer


class CreateDefaultObservatoryLayerTEstCase(TestCase):
    @override_settings(TROPP_OBSERVATORY_LAYER_PK=None)
    def test_create_without_existing_layer(self):
        out = StringIO()
        call_command("create_observatory_layer", name="test", stdout=out)
        self.assertIn("Layer has been created", out.getvalue())

    @override_settings(TROPP_OBSERVATORY_LAYER_PK=99999)
    def test_create_with_nonexisting_configured_layer(self):
        out = StringIO()
        call_command("create_observatory_layer", name="test", stdout=out)
        self.assertIn("does not exists in database", out.getvalue())

    @override_settings(TROPP_OBSERVATORY_LAYER_PK=99999)
    def test_force_create_with_nonexisting_configured_layer(self):
        out = StringIO()
        call_command("create_observatory_layer", name="test", force=True, stdout=out)
        self.assertIn("Layer has been created", out.getvalue())

    def test_force_create_with_existing_configured_layer(self):
        out = StringIO()
        layer = Layer.objects.create(name="tmp", geom_type=GeometryTypes.Point)
        with override_settings(TROPP_OBSERVATORY_LAYER_PK=layer.pk):
            call_command(
                "create_observatory_layer", name="test", force=True, stdout=out
            )
            self.assertIn("A layer already exists", out.getvalue())

    @override_settings(TROPP_OBSERVATORY_LAYER_PK=99999)
    def test_create_with_existing_name(self):
        out = StringIO()
        # create a layer with name test
        Layer.objects.create(name="test", geom_type=GeometryTypes.Point)
        call_command("create_observatory_layer", name="test", force=True, stdout=out)
        self.assertIn("already exists for this name", out.getvalue())

    def test_create_with_right_settings(self):
        out = StringIO()
        # create a layer
        layer = Layer.objects.create(name="tmp", geom_type=GeometryTypes.Point)
        with override_settings(TROPP_OBSERVATORY_LAYER_PK=layer.pk):
            call_command("create_observatory_layer", name="test", stdout=out)
            self.assertIn("A layer already exists", out.getvalue())



import factory
from django.utils import timezone
from factory.django import FileField

from terra_opp.models import Campaign, City, Picture, Theme, Viewpoint


class CampaignFactory(factory.django.DjangoModelFactory):
    owner = factory.SubFactory("terra_accounts.tests.factories.TerraUserFactory")
    assignee = factory.SubFactory("terra_accounts.tests.factories.TerraUserFactory")

    class Meta:
        model = Campaign


class ViewpointFactory(factory.django.DjangoModelFactory):
    point = factory.SubFactory("geostore.tests.factories.FeatureFactory")
    pictures = factory.RelatedFactory(
        "terra_opp.tests.factories.PictureFactory", "viewpoint"
    )
    city = factory.SubFactory("terra_opp.tests.factories.CityFactory")
    themes = factory.RelatedFactory("terra_opp.tests.factories.ThemeFactory")

    class Meta:
        model = Viewpoint


class PictureFactory(factory.django.DjangoModelFactory):
    owner = factory.SubFactory("terra_accounts.tests.factories.TerraUserFactory")
    date = timezone.datetime(2018, 1, 1, tzinfo=timezone.utc)
    file = FileField(from_path="terra_opp/tests/placeholder.jpg")

    class Meta:
        model = Picture


class CityFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = City


class ThemeFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Theme



from io import StringIO

from django.test import TestCase

from terra_opp.helpers import Choices, CustomCsvBuilder


class CSVTestCase(TestCase):
    def test_create_csv_1(self):
        data = {
            "test": "hello-world",
            "value": [
                {
                    "data": {
                        "test1": {
                            "first": {"events": 1, "participants": 1},
                            "second": {"events": 2, "participants": 2},
                            "third": {"events": 1, "participants": 101},
                        }
                    },
                    "name": "Name1",
                },
                {
                    "data": {
                        "test0": {"fourth": {"events": 1, "participants": 1}},
                        "test1": {
                            "first": {"events": 1, "participants": 1},
                            "second": {"events": 8, "participants": 73},
                            "third": {"events": 1, "participants": 101},
                        },
                    },
                    "name": "Name2",
                },
            ],
        }

        csv_builder = CustomCsvBuilder(data)

        output = StringIO()
        csv_builder.create_csv(output)
        list_lines = self.get_columns(output.read())
        self.assertCountEqual(
            list_lines,
            [
                ["test", "hello-world", ""],
                ["value", "", ""],
                ["value_data_test0_fourth_events", "", "1"],
                ["value_data_test0_fourth_participants", "", "1"],
                ["value_data_test1_first_events", "1", "1"],
                ["value_data_test1_first_participants", "1", "1"],
                ["value_data_test1_second_events", "2", "8"],
                ["value_data_test1_second_participants", "2", "73"],
                ["value_data_test1_third_events", "1", "1"],
                ["value_data_test1_third_participants", "101", "101"],
                ["value_name", "Name1", "Name2"],
            ],
        )

    def test_create_csv_2(self):
        data = {"test": "hello-world", "test2": "HELLO-WORLD"}
        csv_builder = CustomCsvBuilder(data)
        output = StringIO()
        csv_builder.create_csv(output)
        list_lines = self.get_columns(output.read())
        self.assertCountEqual(
            list_lines, [["test", "hello-world"], ["test2", "HELLO-WORLD"]]
        )

    def get_columns(self, csv):
        list_csv = []
        tab = csv.split("\r\n")
        tab.pop(-1)
        for line in tab:
            values = line.split(",")
            list_csv.append(values)
        by_column = []
        for id in range(len(list_csv[0])):
            column = []
            for line in list_csv:
                column.append(line[id])
            by_column.append(column)
        return by_column


class ChoicesTests(TestCase):
    def setUp(self):
        self.MY_CHOICES = Choices(
            ("ONE", 1, "One for the money"),
            ("TWO", 2, "Two for the show"),
            ("THREE", 3, "Three to get ready"),
        )
        self.MY_CHOICES.add_subset("ODD", ("ONE", "THREE"))

    """
    Testing the choices
    """

    def test_simple_choice(self):
        self.assertEqual(
            self.MY_CHOICES.CHOICES,
            (
                (1, "One for the money"),
                (2, "Two for the show"),
                (3, "Three to get ready"),
            ),
        )
        self.assertEqual(
            self.MY_CHOICES.CHOICES_DICT,
            {1: "One for the money", 2: "Two for the show", 3: "Three to get ready"},
        )
        self.assertEqual(
            self.MY_CHOICES.REVERTED_CHOICES_DICT,
            {"One for the money": 1, "Three to get ready": 3, "Two for the show": 2},
        )

    def test__contains__(self):
        self.failUnless(self.MY_CHOICES.ONE in self.MY_CHOICES)

    def test__iter__(self):
        self.assertEqual([k for k, v in self.MY_CHOICES], [1, 2, 3])

    def test_unique_values(self):
        self.assertRaises(
            ValueError, Choices, ("TWO", 4, "Deux"), ("FOUR", 4, "Quatre")
        )

    def test_unique_constants(self):
        self.assertRaises(ValueError, Choices, ("TWO", 2, "Deux"), ("TWO", 4, "Quatre"))

    def test_const_choice(self):
        self.assertEqual(
            self.MY_CHOICES.CONST_CHOICES,
            (
                ("ONE", "One for the money"),
                ("TWO", "Two for the show"),
                ("THREE", "Three to get ready"),
            ),
        )

    def test_value_to_const(self):
        self.assertEqual(
            self.MY_CHOICES.VALUE_TO_CONST, {1: "ONE", 2: "TWO", 3: "THREE"}
        )

    def test_add_should_add_in_correct_order(self):
        SOME_CHOICES = Choices(
            ("ONE", 1, "One"),
            ("TWO", 2, "Two"),
        )
        OTHER_CHOICES = Choices(
            ("THREE", 3, "Three"),
            ("FOUR", 4, "Four"),
        )
        # Adding a choices to choices
        tup = SOME_CHOICES + OTHER_CHOICES
        self.assertEqual(tup, ((1, "One"), (2, "Two"), (3, "Three"), (4, "Four")))

        # Adding a tuple to choices
        tup = SOME_CHOICES + ((3, "Three"), (4, "Four"))
        self.assertEqual(tup, ((1, "One"), (2, "Two"), (3, "Three"), (4, "Four")))

        """Adding a choices to tuple => do not work; is it possible to
           emulate it?
            tup = ((1, 'One'), (2, 'Two')) + OTHER_CHOICES
            self.assertEqual(tup, ((1, 'One'), (2, 'Two'),
                                   (3, 'Three'), (4, 'Four')))
        """

    def test_retrocompatibility(self):
        MY_CHOICES = Choices(("TWO", 2, "Deux"), ("FOUR", 4, "Quatre"), name="EVEN")
        MY_CHOICES.add_choices(
            "ODD",
            ("ONE", 1, "Un"),
            ("THREE", 3, "Trois"),
        )
        self.assertEqual(
            MY_CHOICES.CHOICES, ((2, "Deux"), (4, "Quatre"), (1, "Un"), (3, "Trois"))
        )
        self.assertEqual(MY_CHOICES.ODD, ((1, "Un"), (3, "Trois")))
        self.assertEqual(MY_CHOICES.EVEN, ((2, "Deux"), (4, "Quatre")))


class SubsetTests(TestCase):
    def setUp(self):
        self.MY_CHOICES = Choices(
            ("ONE", 1, "One for the money"),
            ("TWO", 2, "Two for the show"),
            ("THREE", 3, "Three to get ready"),
        )
        self.MY_CHOICES.add_subset("ODD", ("ONE", "THREE"))
        self.MY_CHOICES.add_subset("ODD_BIS", ("ONE", "THREE"))

    def test_basic(self):
        self.assertEqual(
            self.MY_CHOICES.ODD, ((1, "One for the money"), (3, "Three to get ready"))
        )

    def test__contains__(self):
        self.failUnless(self.MY_CHOICES.ONE in self.MY_CHOICES.ODD)

    def test__eq__(self):
        self.assertEqual(
            self.MY_CHOICES.ODD, ((1, "One for the money"), (3, "Three to get ready"))
        )
        self.assertEqual(self.MY_CHOICES.ODD, self.MY_CHOICES.ODD_BIS)



# Generated by Django 2.2.5 on 2020-06-30 13:15

from django.db import migrations, models
import django.db.models.deletion
from django.test import RequestFactory
from terra_opp.point_utilities import update_point_properties


def set_city_and_themes_from_properties(app, schema_editor):
    Viewpoint = app.get_model("terra_opp", "Viewpoint")
    City = app.get_model("terra_opp", "City")
    Theme = app.get_model("terra_opp", "Theme")
    for viewpoint in Viewpoint.objects.all():
        city, created = City.objects.get_or_create(
            label=viewpoint.properties["commune"]
        )
        viewpoint.city = city
        del viewpoint.properties["commune"]
        for theme in viewpoint.properties["themes"]:
            theme, created = Theme.objects.get_or_create(label=theme)
            viewpoint.themes.add(theme)
        del viewpoint.properties["themes"]
        viewpoint.save()
        update_point_properties(viewpoint, RequestFactory().get("/"))


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0002_remove_picture_remarks"),
    ]

    operations = [
        migrations.CreateModel(
            name="City",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("label", models.CharField(max_length=100, verbose_name="Label")),
            ],
            options={
                "abstract": False,
                "verbose_name_plural": "Cities",
            },
        ),
        migrations.CreateModel(
            name="Theme",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("label", models.CharField(max_length=100, verbose_name="Label")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="viewpoint",
            name="city",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="viewpoints",
                to="terra_opp.City",
            ),
        ),
        migrations.AddField(
            model_name="viewpoint",
            name="themes",
            field=models.ManyToManyField(
                related_name="viewpoints", to="terra_opp.Theme"
            ),
        ),
        migrations.RunPython(
            set_city_and_themes_from_properties, reverse_code=migrations.RunPython.noop
        ),
    ]



from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.gis.db import models

try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField
from django.utils.translation import ugettext_lazy as _
from versatileimagefield.fields import VersatileImageField

from terra_settings.mixins import BaseUpdatableModel
from geostore.models import Feature
from .settings import TROPP_STATES as DEFAULT_TROPP_STATES


class BaseLabelModel(BaseUpdatableModel):
    label = models.CharField(_("Label"), max_length=100)

    class Meta:
        abstract = True

    def __str__(self):
        return self.label


class ViewpointsManager(models.Manager):
    def with_accepted_pictures(self):
        return (
            super()
            .get_queryset()
            .filter(
                pictures__state=settings.TROPP_STATES.ACCEPTED,
            )
            .distinct()
        )


class City(BaseLabelModel):
    class Meta:
        verbose_name_plural = _("Cities")


class Theme(BaseLabelModel):
    pass


class Viewpoint(BaseLabelModel):
    point = models.ForeignKey(
        Feature,
        on_delete=models.CASCADE,
        related_name="points",
    )
    city = models.ForeignKey(
        City,
        on_delete=models.CASCADE,
        related_name="viewpoints",
        null=True,
    )
    themes = models.ManyToManyField(
        Theme,
        related_name="viewpoints",
    )
    properties = JSONField(_("Properties"), default=dict, blank=True)
    related = GenericRelation("datastore.RelatedDocument")

    objects = ViewpointsManager()

    @property
    def status(self):
        """
        Return the status of this viewpoint for a campaign
        :param self:
        :return: string (missing, draft, submitted, accepted)
        """
        # Get only pictures created for the campaign
        picture = self.pictures.latest()
        STATES = settings.TROPP_STATES
        if picture.created_at < self.created_at:
            return STATES.CHOICES_DICT[STATES.MISSING]
        return STATES.CHOICES_DICT[picture.state]

    @property
    def ordered_pics(self):
        if hasattr(self, "_ordered_pics"):
            # if _ordered_pics set by prefetch on qs, get it
            return self._ordered_pics
        else:
            # not prefetch, compute
            return self.pictures.order_by("-created_at")

    @property
    def picture(self):
        pics = self.ordered_pics
        return pics[0].file if pics else None

    class Meta:
        permissions = (("can_download_pdf", "Is able to download a pdf document"),)
        ordering = ["-created_at"]


class Campaign(BaseLabelModel):
    viewpoints = models.ManyToManyField(
        Viewpoint,
        related_name="campaigns",
    )
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_("Owner"),
        related_name="campaigns",
    )
    assignee = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_("Assigned to"),
        related_name="assigned_campaigns",
    )

    @property
    def statistics(self):
        queryset = self.viewpoints.annotate(
            missing=models.Count("pk", filter=models.Q(pictures__isnull=True)),
            pending=models.Count(
                "pictures", filter=models.Q(pictures__state=settings.TROPP_STATES.DRAFT)
            ),
            refused=models.Count(
                "pictures",
                filter=models.Q(pictures__state=settings.TROPP_STATES.REFUSED),
            ),
        ).values("missing", "pending", "refused")
        try:
            return queryset[0]
        except IndexError:
            return {"missing": 0, "pending": 0, "refused": 0}

    @property
    def status(self):
        return not self.viewpoints.exclude(
            pictures__state=settings.TROPP_STATES.ACCEPTED,
        ).exists()

    class Meta:
        permissions = (("manage_all_campaigns", "Can manage all campaigns"),)
        ordering = ["-created_at"]


class Picture(BaseUpdatableModel):
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_("Owner"),
        related_name="pictures",
    )
    viewpoint = models.ForeignKey(
        Viewpoint,
        on_delete=models.CASCADE,
        related_name="pictures",
    )
    # States may be : draft, metadata_ok (submitted), accepted, refused
    state = models.IntegerField(
        _("State"),
        default=getattr(settings, "TROPP_STATES", DEFAULT_TROPP_STATES).DRAFT,
    )

    properties = JSONField(_("Properties"), default=dict, blank=True)
    file = VersatileImageField(_("File"))

    # Different from created_at which is the upload date
    date = models.DateTimeField(_("Date"))

    class Meta:
        permissions = (
            ("change_state_picture", "Is able to change the picture " "state"),
        )
        # It's our main way of sorting pictures, so it better be indexed
        indexes = [
            models.Index(fields=["viewpoint", "date"]),
        ]
        get_latest_by = "date"
        ordering = ["-date"]

    def save(self, *args, **kwargs):
        if not settings.TROPP_PICTURES_STATES_WORKFLOW:
            self.state = settings.TROPP_STATES.ACCEPTED
        super().save(*args, **kwargs)



from django.conf import settings
from django.shortcuts import resolve_url
from django.test import override_settings
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from terra_accounts.tests.factories import TerraUserFactory
from terra_opp.tests.factories import CampaignFactory, ViewpointFactory
from terra_opp.tests.mixins import TestPermissionsMixin


class CampaignTestCase(TestPermissionsMixin, APITestCase):
    def setUp(self):
        self.photograph = TerraUserFactory()
        self.user = TerraUserFactory()
        self._set_permissions(
            [
                "manage_all_campaigns",
                "add_campaign",
            ]
        )

    def test_list_campaign(self):
        viewpoint2 = ViewpointFactory()
        viewpoint = ViewpointFactory()
        campaign = CampaignFactory(assignee=self.photograph)
        campaign.viewpoints.set([viewpoint, viewpoint2])

        campaign_other = CampaignFactory()  # campaign for other photograph

        list_url = reverse("terra_opp:campaign-list")
        campaign_url = resolve_url("terra_opp:campaign-detail", pk=campaign.pk)
        campaign_other_url = resolve_url(
            "terra_opp:campaign-detail", pk=campaign_other.pk
        )

        # First we try as anonymous
        self.assertIn(
            self.client.get(list_url).status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )
        self.assertIn(
            self.client.get(campaign_url).status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )
        self.assertIn(
            self.client.get(campaign_other_url).status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )

        # Then with no rights
        self.client.force_authenticate(user=self.photograph)
        response = self.client.get(list_url)
        self.assertEqual(status.HTTP_200_OK, response.status_code)
        self.assertEqual(1, response.data.get("count"))
        self.assertIn(
            viewpoint.pictures.first().file.url,
            response.data.get("results")[0].get("picture").get("original"),
        )

        response = self.client.get(campaign_url)
        self.assertEqual(status.HTTP_200_OK, response.status_code)
        self.assertIn(
            viewpoint.pictures.first().file.url,
            response.data.get("viewpoints")[0].get("picture").get("original"),
        )
        self.assertEqual(
            status.HTTP_403_FORBIDDEN, self.client.get(campaign_other_url).status_code
        )

        # Then as admin
        self.client.force_authenticate(user=self.user)
        response = self.client.get(list_url)
        self.assertEqual(status.HTTP_200_OK, response.status_code)
        self.assertEqual(2, response.data.get("count"))

        self.assertEqual(status.HTTP_200_OK, self.client.get(campaign_url).status_code)
        self.assertEqual(
            status.HTTP_200_OK, self.client.get(campaign_other_url).status_code
        )

    @override_settings(TROPP_PICTURES_STATES_WORKFLOW=True)
    def test_get_campaign(self):
        campaign = CampaignFactory(assignee=self.photograph)
        campaign_url = resolve_url("terra_opp:campaign-detail", pk=campaign.pk)
        list_url = resolve_url("terra_opp:campaign-list")
        self.client.force_authenticate(user=self.photograph)

        viewpoint = ViewpointFactory()
        viewpoint.pictures.all().delete()
        campaign.viewpoints.set([viewpoint])
        response = self.client.get(campaign_url)
        self.assertEqual("Missing", response.data.get("viewpoints")[0].get("status"))
        response = self.client.get(list_url)
        self.assertEqual(
            1, response.data.get("results")[0].get("statistics").get("missing")
        )

        viewpoint = ViewpointFactory()
        campaign.viewpoints.set([viewpoint])
        response = self.client.get(campaign_url)

        self.assertEqual("Draft", response.data.get("viewpoints")[0].get("status"))
        response = self.client.get(list_url)
        self.assertEqual(
            1, response.data.get("results")[0].get("statistics").get("pending")
        )

    @override_settings(TROPP_PICTURES_STATES_WORKFLOW=True)
    def test_search_campaign(self):
        campaign = CampaignFactory(assignee=self.photograph)
        list_url = resolve_url("terra_opp:campaign-list")
        self.client.force_authenticate(user=self.user)

        viewpoint = ViewpointFactory()
        campaign.viewpoints.set([viewpoint])
        # Picture state is draft
        response = self.client.get(list_url, {"status": 0})
        self.assertEqual(1, response.data.get("count"))
        response = self.client.get(list_url, {"status": 1})
        self.assertEqual(0, response.data.get("count"))
        response = self.client.get(list_url, {"picture_status": 100})
        self.assertEqual(1, response.data.get("count"))
        response = self.client.get(list_url, {"picture_status": 200})
        self.assertEqual(0, response.data.get("count"))

        viewpoint.pictures.update(state=settings.TROPP_STATES.ACCEPTED)
        response = self.client.get(list_url, {"status": 0})
        self.assertEqual(0, response.data.get("count"))
        response = self.client.get(list_url, {"status": 1})
        self.assertEqual(1, response.data.get("count"))
        response = self.client.get(list_url, {"picture_status": 200})
        self.assertEqual(0, response.data.get("count"))
        response = self.client.get(list_url, {"picture_status": 300})
        self.assertEqual(1, response.data.get("count"))

    def test_post_campaign(self):
        data = {
            "label": "My campaign",
            "assignee": self.photograph.pk,
            "viewpoints": [ViewpointFactory().pk, ViewpointFactory().pk],
        }

        self.client.force_authenticate(user=self.photograph)
        response = self.client.post(reverse("terra_opp:campaign-list"), data)
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

        self.client.force_authenticate(user=self.user)
        response = self.client.post(reverse("terra_opp:campaign-list"), data)
        self.assertEqual(status.HTTP_201_CREATED, response.status_code)



#!/usr/bin/env python

import os
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

README = open(os.path.join(HERE, 'README.md')).read()
CHANGES = open(os.path.join(HERE, 'CHANGES.md')).read()

setup(
    name='terra-opp',
    version=open(os.path.join(HERE, 'terra_opp', 'VERSION.md')).read().strip(),
    include_package_data=True,
    author="Makina Corpus",
    author_email="terralego-pypi@makina-corpus.com",
    description='Observatoire Photographique des Paysages',
    long_description=README + '\n\n' + CHANGES,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    url='https://github.com/Terralego/terra-opp.git',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    install_requires=[
        'Django>=2.2',
        'djangorestframework',
        'coreapi',
        'coreschema',
        'djangorestframework-gis',
        'django-url-filter',
        'django-geostore>=0.3.15',
        'django-datastore>=0.1.1,<0.2',
        'django-terra-accounts>=0.3.11',
        "django-terra-settings",
        "psycopg2>=2.7",
        "django-versatileimagefield",
        "weasyprint",
        "django-filter",
    ],
    extras_require={
        'dev': [
            'factory-boy',
            'flake8',
            'black',
            'coverage',
        ]
    }
)



from pprint import pformat

import coreapi
import coreschema
from django.conf import settings
from django.contrib.postgres.fields.jsonb import KeyTextTransform
from django.db.models import Q
from django_filters import FilterSet
from django_filters.rest_framework import filters
from rest_framework.exceptions import ValidationError
from rest_framework.filters import BaseFilterBackend
from url_filter.integrations.drf import DjangoFilterBackend

from .models import Viewpoint


class CampaignFilterBackend(BaseFilterBackend):
    """
    Filters for campaigns
    """

    def get_schema_fields(self, view):
        super().get_schema_fields(view)
        choices = {
            settings.TROPP_STATES.DRAFT: "Incomplete metadata",
            settings.TROPP_STATES.SUBMITTED: "Pending validation",
            settings.TROPP_STATES.REFUSED: "Refused",
            settings.TROPP_STATES.ACCEPTED: "Validated",
        }
        return [
            coreapi.Field(
                name="status",
                required=False,
                location="query",
                schema=coreschema.Boolean(
                    title="Campaign status",
                    description="0 for closed campaign, 1 for ongoing campaign",
                ),
            ),
            coreapi.Field(
                name="picture_status",
                required=False,
                location="query",
                schema=coreschema.Enum(
                    choices,
                    description=str(pformat(choices)),
                    title="Picture status",
                ),
            ),
        ]

    def filter_queryset(self, request, queryset, view):
        STATES = settings.TROPP_STATES
        status = request.GET.get("status", None)
        if status is not None:
            try:
                status = bool(int(status))
            except ValueError:
                raise ValidationError
            if status:
                queryset = queryset.filter(viewpoints__pictures__state=STATES.ACCEPTED)
            else:
                queryset = queryset.exclude(
                    Q(viewpoints__pictures__state=STATES.ACCEPTED)
                )

        picture_status = request.GET.get("picture_status", None)
        if picture_status is not None:
            try:
                picture_status = int(picture_status)
                assert picture_status in STATES.CHOICES_DICT
            except (AssertionError, ValueError):
                raise ValidationError
            queryset = queryset.filter(viewpoints__pictures__state=picture_status)

        return queryset


class JsonFilterBackend(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        for key, field in settings.TROPP_SEARCHABLE_PROPERTIES.items():
            search_key = f'properties__{field["json_key"]}'
            if field["type"] == "many":
                search_item = request.GET.getlist(f"{search_key}")
            else:
                search_item = request.GET.get(search_key)
            if search_item:
                if field["type"] == "text":
                    # Full text search
                    queryset = queryset.annotate(
                        **{key: KeyTextTransform(field["json_key"], "properties")}
                    ).filter(**{f"{key}__icontains": search_item})
                else:
                    # Search on element
                    queryset = queryset.filter(
                        **{f"{search_key}__contains": search_item}
                    )
        return queryset

    def get_schema_fields(self, view):
        fields = []
        for key, field in settings.TROPP_SEARCHABLE_PROPERTIES.items():
            if field["type"] == "many":
                klass = coreschema.Array
            else:
                klass = coreschema.String
            description = f"{key.capitalize()} property ({field['type']})"
            fields.append(
                coreapi.Field(
                    name=f'properties__{field["json_key"]}',
                    required=False,
                    location="query",
                    schema=klass(
                        title=f"{key.capitalize()} property",
                        description=description,
                    ),
                )
            )
        return super().get_schema_fields(view)


class ViewpointFilterSet(FilterSet):
    city = filters.CharFilter(field_name="city__label", lookup_expr="exact")
    themes = filters.CharFilter(field_name="themes__label", lookup_expr="exact")
    date_from = filters.DateFilter(field_name="pictures__date", lookup_expr="gte")
    date_to = filters.DateFilter(field_name="pictures__date", lookup_expr="lte")

    class Meta:
        model = Viewpoint
        fields = ["city", "themes", "date_from", "date_to"]


class SchemaAwareDjangoFilterBackend(DjangoFilterBackend):
    def get_schema_fields(self, view):
        """
        Get coreapi filter definitions

        Returns all schemas defined in filter_fields_schema attribute.
        """
        super().get_schema_fields(view)
        return getattr(view, "filter_fields_schema", [])



import operator
from functools import reduce

import coreapi
import coreschema
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.postgres.fields.jsonb import KeyTransform
from django.core.cache import cache
from django.db.models import Prefetch
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import permissions, viewsets
from rest_framework.decorators import action
from rest_framework.filters import SearchFilter
from rest_framework.response import Response

from .filters import (
    CampaignFilterBackend,
    JsonFilterBackend,
    ViewpointFilterSet,
    SchemaAwareDjangoFilterBackend,
)
from .models import Campaign, City, Picture, Theme, Viewpoint
from .pagination import RestPageNumberPagination
from .point_utilities import remove_point_thumbnail, update_point_properties
from .renderers import PdfRenderer, ZipRenderer
from .serializers import (
    CampaignSerializer,
    DetailAuthenticatedCampaignNestedSerializer,
    DetailCampaignNestedSerializer,
    ListCampaignNestedSerializer,
    PictureSerializer,
    SimpleAuthenticatedViewpointSerializer,
    SimpleViewpointSerializer,
    ViewpointSerializerWithPicture,
    PhotographSerializer,
)


class ViewpointViewSet(viewsets.ModelViewSet):
    serializer_class = ViewpointSerializerWithPicture
    permission_classes = [
        permissions.DjangoModelPermissionsOrAnonReadOnly,
    ]
    filter_backends = (
        SearchFilter,
        SchemaAwareDjangoFilterBackend,
        JsonFilterBackend,
        DjangoFilterBackend,
    )
    filterset_class = ViewpointFilterSet
    filter_fields_schema = [
        coreapi.Field(
            name="pictures__id",
            required=False,
            location="query",
            schema=coreschema.Integer(
                title="Picture id",
                description="Picture id to filter on",
            ),
        ),
        coreapi.Field(
            name="pictures__owner__uuid",
            required=False,
            location="query",
            schema=coreschema.Integer(
                title="Photographer uuid",
                description="Photographer uuid to filter on",
            ),
        ),
    ]
    filter_fields = ["pictures"]
    search_fields = ("label",)
    pagination_class = RestPageNumberPagination
    template_name = "terra_opp/viewpoint_pdf.html"

    def perform_create(self, serializer):
        serializer.save()
        update_point_properties(serializer.instance, self.request)

    def perform_update(self, serializer):
        serializer.save()
        update_point_properties(serializer.instance, self.request)

    def perform_destroy(self, instance):
        instance.point.delete()
        instance.delete()
        # FIXME This may be better if done directly in geostore
        # FIXME Try to be more precise and delete only the related feature's tile in cache
        cache.delete("tile_cache_*")  # delete all the cached tiles

    def filter_queryset(self, queryset):
        # We must reorder the queryset here because initial filtering in
        # viewpoint model is not done right see
        # https://github.com/encode/django-rest-framework/issues/1717
        return super().filter_queryset(queryset).order_by("-created_at")

    def get_queryset(self):
        qs = Viewpoint.objects.with_accepted_pictures()
        if self.request.user.is_authenticated:
            qs = Viewpoint.objects.all().distinct()
        pictures_qs = Picture.objects.order_by("-created_at")
        return qs.select_related("point", "city").prefetch_related(
            "pictures",
            "related",
            Prefetch("pictures", queryset=pictures_qs, to_attr="_ordered_pics"),
            "themes",
        )

    def get_serializer_class(self):
        if self.action == "list":
            if self.request.user.is_anonymous:
                return SimpleViewpointSerializer
            return SimpleAuthenticatedViewpointSerializer
        return ViewpointSerializerWithPicture

    @action(detail=False)
    def filters(self, request, *args, **kwargs):
        filter_values = {}
        for key, field in settings.TROPP_SEARCHABLE_PROPERTIES.items():
            data = None
            transform = KeyTransform(field["json_key"], "properties")
            queryset = (
                Viewpoint.objects.annotate(**{key: transform})
                .exclude(**{f"{key}__isnull": True})
                .values_list(key, flat=True)
            )
            if field["type"] == "single":
                # Dedupe and sort with SQL
                data = queryset.order_by(key).distinct(key)
            elif field["type"] == "many":
                # Dedupe and sort programmatically
                data = list(queryset)
                if data:
                    data = set(reduce(operator.concat, data))
                    data = sorted(data, key=str.lower)
            if data is not None:
                filter_values[key] = data

        filter_values["photographers"] = PhotographSerializer(
            get_user_model().objects.filter(pictures__isnull=False).distinct(),
            many=True,
        ).data

        # FIXME We may want to set all cities and themes as uniques directly in the model?
        filter_values["cities"] = (
            City.objects.exclude(label__isnull=True)
            .exclude(label__exact="")
            .distinct()
            .order_by(
                "label",
            )
            .values_list(
                "label",
                flat=True,
            )
        )
        filter_values["themes"] = (
            Theme.objects.exclude(label__isnull=True)
            .exclude(
                label__exact="",
            )
            .distinct()
            .order_by(
                "label",
            )
            .values_list(
                "label",
                flat=True,
            )
        )

        return Response(filter_values)

    @method_decorator(cache_page(60 * 5))
    @action(
        detail=True,
        methods=[
            "get",
        ],
        renderer_classes=[ZipRenderer],
        url_path="zip-pictures",
    )
    def zip_pictures(self, request, *args, **kwargs):
        qs = (
            self.get_object()
            .pictures.filter(
                state__gte=settings.TROPP_STATES.ACCEPTED,
            )
            .only("file")
        )
        return Response([p.file for p in qs])

    @method_decorator(cache_page(60 * 5))
    @action(
        detail=True,
        methods=[
            "get",
        ],
        renderer_classes=[PdfRenderer],
    )
    def pdf(self, request, *args, **kwargs):
        properties_set = settings.TROPP_VIEWPOINT_PROPERTIES_SET["pdf"]
        return Response(
            {
                "viewpoint": self.get_object(),
                "properties_set": properties_set,
            }
        )


class CampaignViewSet(viewsets.ModelViewSet):
    queryset = Campaign.objects.all()
    permission_classes = [permissions.DjangoModelPermissions]
    http_method_names = ["get", "post", "put", "delete", "options"]
    filter_backends = (CampaignFilterBackend, SearchFilter)
    search_fields = ("label",)
    pagination_class = RestPageNumberPagination

    def get_queryset(self):
        # Filter only on assigned campaigns for photographs
        user = self.request.user
        pictures_qs = Picture.objects.order_by("-created_at")
        qs = (
            super()
            .get_queryset()
            .prefetch_related(
                Prefetch(
                    "viewpoints__pictures", queryset=pictures_qs, to_attr="ordered_pics"
                )
            )
        )
        if self.action == "list" and not user.has_perm(
            "terra_opp.manage_all_campaigns"
        ):
            return qs.filter(assignee=user)
        return qs

    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

    def check_object_permissions(self, request, obj: Campaign):
        # Prevent acting on unassigned campaigns for photographs
        has_perm = request.user.has_perm("terra_opp.manage_all_campaigns")
        if not has_perm and obj.assignee != request.user:
            self.permission_denied(request)
        super().check_object_permissions(request, obj)

    def get_serializer_class(self):
        if self.action == "list":
            return ListCampaignNestedSerializer
        if self.action == "retrieve":
            if self.request.user.is_anonymous:
                return DetailCampaignNestedSerializer
            return DetailAuthenticatedCampaignNestedSerializer
        return CampaignSerializer


class PictureViewSet(viewsets.ModelViewSet):
    queryset = Picture.objects.all()
    serializer_class = PictureSerializer
    permission_classes = [permissions.DjangoModelPermissions]
    pagination_class = RestPageNumberPagination

    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)
        update_point_properties(serializer.instance.viewpoint, self.request)

    def perform_update(self, serializer):
        serializer.save()
        update_point_properties(serializer.instance.viewpoint, self.request)

    def perform_destroy(self, instance):
        remove_point_thumbnail(instance, self.request)
        instance.delete()
